from .glvq import GlvqModel
from .grlvq import GrlvqModel
from .gmlvq import GmlvqModel
from .lgmlvq import LgmlvqModel
from .grmlvq import GrmlvqModel
from .rslvq import RslvqModel
from .mrslvq import MrslvqModel
from .lmrslvq import LmrslvqModel

__all__ = ['GlvqModel', 'GrlvqModel', 'GmlvqModel', 'GrmlvqModel', 'LgmlvqModel',
           'RslvqModel', 'MrslvqModel', 'LmrslvqModel']
__version__ = '1.1.0'
